/*  Author: Zhi Xin 
 *  Data: 11/16/2011
 *  Location: IST, PSU
 *  Function: 1.generate the mapping relation between glibc functions and system calls by FILE "database"
 *	      2.generate the system calls distriution for target program by STANDARD OUTPUT
 */


#include <stdio.h>
#include <stdlib.h>
#include <search.h>
#include <string.h>

#define MAXFUNCNAME 50		// some function's name longer than 30
#define MAXFUNCNUM 10000	


typedef struct subFunc {
	char name[MAXFUNCNAME];
	struct subFunc* next;
} SubFunc;

typedef struct sysCall {
	int syscallnum;
	struct sysCall* next;	
} Syscall;


typedef struct {
	char  name[MAXFUNCNAME];	// the name of current function
	SubFunc* subfunc;  		// subfunctions called by current function;
	Syscall* syscall;      		// I assume that a library function won't directly call system calls more than 10 times.
} FuncNode;


//////////////////////////

void
print_FuncNode(FuncNode* pfnode)
{
	// print the main function name 
	if(pfnode == NULL)
	{
		//printf("EMpty node!!!\n");
		return;
	}

	printf("Func %s", pfnode->name);
	
	// print the sub function names
	SubFunc* ptemSub = pfnode->subfunc;

	{
		printf("  Sub %s",ptemSub->name);
		ptemSub = ptemSub->next;
	}

	// print the direct system call numbers
	Syscall* ptemSys = pfnode->syscall;
	while(ptemSys != NULL)
	{
		printf("  Sys %d\n",ptemSys->syscallnum);
		ptemSys = ptemSys->next;
	}
}

// recursively browse all the mapping relation between function all 
// and direct system call
void
print_FuncNode_syscall(char* line, FILE* output, int level)
{
     	ENTRY e, *ey;

     	e.key = strndup(line, strlen(line)-1);
     	ey = hsearch(e, FIND);
                
	if(ey == NULL || level > 5)
        {
             	//printf("Emopty key\n");
         } else {
             	//printf("Succeful\n");
   		// print the main function name 
		FuncNode* pfnode = ((FuncNode*)ey->data);
		if(level == 0)
        		fprintf(output, "%s", pfnode->name);

        	// print the direct system call numbers
        	Syscall* ptemSys = pfnode->syscall;
        	while(ptemSys != NULL)
        	{
                	fprintf(output, "%d ",ptemSys->syscallnum);
                	ptemSys = ptemSys->next;
        	}

		// recursively call this function 
		SubFunc* ptemSub = pfnode->subfunc;
        	while(ptemSub != NULL)
        	{
                	//printf("  Sub %s",ptemSub->name);
			// recursive case !!!
			if(strcmp(pfnode->name, ptemSub->name) != 0)
			{
				print_FuncNode_syscall(ptemSub->name, output, level+1);
			} else {
				printf("Recursive Error!\n");
				return;
			}
                	ptemSub = ptemSub->next;
        	}
	}
}


void
print_libsyscall_mapping(FILE* finput, FILE* output)
{
  	char* line = malloc(MAXFUNCNAME);
        int strmax = MAXFUNCNAME;
	int level = 0;

 	while((getline(&line, &strmax, finput)) != -1)
        {
		level = 0;	
		if(line != NULL)
		{
			print_FuncNode_syscall(line, output, level);
			fprintf(output, "\n");
		}
	}
        free(line);
}


// just a test for three popular system calls
void
print_test_hashtable(void)
{
    	char* names[] = {"access", "sigsuspend", "uname"};
    	ENTRY e, *ey;
	
	int j;
	for(j = 0; j < 3; j++)
	{
		printf("name: %s\n", names[j]);
	}

	int i;
    	for (i = 0; i < 3; i++)
   	{
    		e.key = names[i];
		ey = hsearch(e, FIND);
		if(ey == NULL)
		{
			printf("Emopty key\n");
			break;
		}
		printf("Succeful\n");
		print_FuncNode(ey->data);
	}
}

void
count_syscalls(char* line, int* syscalls, int level)
{
	ENTRY e, *ey;

	e.key = strndup(line, strlen(line)-1);
        ey = hsearch(e, FIND);

        if(ey == NULL || level > 5)
        {
                //printf("Emopty key\n");
         } else {
                //printf("Succeful\n");
                // print the main function name 
                FuncNode* pfnode = ((FuncNode*)ey->data);
                //if(level == 0)
                //        fprintf(output, "%s", pfnode->name);
                // print the direct system call numbers
                Syscall* ptemSys = pfnode->syscall;
                while(ptemSys != NULL)
                {
                        //fprintf(output, "%d ",ptemSys->syscallnum);
			// count the number
			if(ptemSys->syscallnum >= 0 && ptemSys->syscallnum < 300)
				syscalls[ptemSys->syscallnum]++;
			// sometimes, there is number error. But I don't know why.
			//else 
                        //        printf("Syscall number Error!\n");
                        ptemSys = ptemSys->next;
                }

                // recursively call this function 
                SubFunc* ptemSub = pfnode->subfunc;
                while(ptemSub != NULL)
                {
                        //printf("  Sub %s",ptemSub->name);
                        // recursive case !!!
                        if(strcmp(pfnode->name, ptemSub->name) != 0)
                        {
                                count_syscalls(ptemSub->name, syscalls, level+1);
                        } else {
                                printf("Recursive Error!\n");
                                return;
                        }
                        ptemSub = ptemSub->next;
                }
        }


}

// output the system call statistic information 
// directly to standard output
void
print_target_syscall_distribution(FILE* target)
{
        char* line = malloc(MAXFUNCNAME);
        int strmax = MAXFUNCNAME;
        int level = 0;

	int syscalls[300] = {0};
	memset(syscalls, 0, 300);

        while((getline(&line, &strmax, target)) != -1)
        {
                level = 0;
                if(line != NULL)
                {
                        count_syscalls(line, syscalls, level);
                        //fprintf(output, "\n");
                }
        }
        free(line);

	// after count, let's print
	int i = 1;
	for(; i < 300; i++)
	{
		// index 1 match system call 1
		//if(syscalls[i] != 0)
		//fprintf(sys, "%d\n",syscalls[i]);
		printf("%d\n",syscalls[i]);
	}
	fflush(stdout);
}

// main analysis part
int start_analysis(FILE* input)
{
	char* line = NULL;
	int strmax = 100;
	int sum = 0;
	char* substr = NULL;
	line = (char*) malloc(strmax);

	while((getline(&line, &strmax, input)) != -1)
	{
		sum++;
		//printf("[%d] %s", sum, line);
		
		substr = strtok(line, " ");
		switch(substr[0])
		{
			// new funcion
			case 'F': 
				substr = strtok(NULL, " ");
				FuncNode* pcurNode = malloc(sizeof(FuncNode));
				if(pcurNode == NULL)
				{
					printf("out of memory!!\n");
					return -1;
				}
				// initialize struct
				memset(pcurNode, 0, sizeof(FuncNode));
				pcurNode->subfunc = NULL;
				pcurNode->syscall = NULL;

				if(strlen(substr) > 0)
				{
					memcpy(pcurNode->name, substr, strlen(substr));
					//print_FuncNode(pcurNode);
					//printf("name is %s", pcurNode->name);
				}
				//printf("--------1-------\n");
				ENTRY e, *ey;
				memset(&e, 0 , sizeof(ENTRY));
			  	//printf("%s\n", pcurNode->name);
				//e.key = strndup(pcurNode->name,7);
			  	//printf("length: %d\n", strlen(pcurNode->name));
				//printf("%s",strndup(pcurNode->name, strlen(pcurNode->name)));
				e.key = strndup(pcurNode->name, strlen(pcurNode->name)-1);
				//e.key = pcurNode->name;
				//e.key = "access";
				//printf("entry name %s",e.key);
				e.data = pcurNode;
				ey = hsearch(e, ENTER);
				if( ey == NULL)
				{
					printf("entry failed!!!\n");
				}
				//printf("--------1.5-------\n");

				/*if( substr != NULL)
				{
					printf("Find a Func called %s", substr);
				}*/
				break;
			case 'C':
				substr = strtok(NULL, " ");
				SubFunc* pcurSub = malloc(sizeof(SubFunc));

				if(pcurSub == NULL)
				{
					printf("out of memory!!\n");
					return -1;
				}
				memset(pcurSub, 0, sizeof(SubFunc));
				pcurSub->next = NULL;

				if(strlen(substr) > 0)
				{
					memcpy(pcurSub->name, substr, strlen(substr));
				}
				//printf("--------2-------\n");

				// find the lastest item in linked list
				if(pcurNode->subfunc == NULL)
				{
					pcurNode->subfunc = pcurSub;
				} else {
					SubFunc* ptemSub = pcurNode->subfunc;
					while(ptemSub->next != NULL)
					{
						ptemSub = ptemSub->next;
					}
					ptemSub->next = pcurSub;
				}
				//printf("%s",pcurNode->subfunc->name);
				//printf("--------3-------\n");
				/*if( substr != NULL)
				{
					printf("Find a Sub func called %s", substr);
				}*/
				break;
			case 'S':
				substr = strtok(NULL, " ");
				Syscall* pcurSys = malloc(sizeof(Syscall));

				if(pcurSys == NULL)
				{
					printf("out of memory!!\n");
					return -1;
				}
				memset(pcurSys, 0, sizeof(Syscall));
				pcurSys->next = NULL;
				
				pcurSys->syscallnum = atoi(substr);				
				//printf("--------4-------\n");
				// find the lastest item in linked list
				if(pcurNode->syscall == NULL)
				{
					pcurNode->syscall = pcurSys;
				} else {
					Syscall* ptemSys = pcurNode->syscall;
					while(ptemSys->next != NULL)
					{
						ptemSys = ptemSys->next;
					}
					ptemSys->next = pcurSys;
				}				
	
				//printf("--------5-------\n");
				/*if( substr != NULL)
				{
					printf("Find a Syscall called %s", substr);
				}*/
				break;
			default:
				printf("No match!!!!");
		}

	}
	
	// clean the room
	free(line);
	return 0;
}

#define QMAXPATH 4096 

char* INPUT_FILE = "/home/zxin/Desktop/syscalblock/static_analysis/accountant/libc_mapping"; // input: the essential input file of program: call tree in the glibc
char FUNC_NAMES[QMAXPATH] = {0};  // input: the library function call names in the glibc, xxx_fun
char TARGET_FILE[QMAXPATH] = {0};	 // input: the target program's library function call list, xxx_ext
//char SYSCALLS_FILE[QMAXPATH] = {0}; // output with option s : the API of collection of target program, syscalls_xxx_2
char* OUTPUT_FILE = "database";  // output with option m : the mapping file recording the mapping between glibc function calls and system calls


int main(int argc, char* argv[]) {
	char* option = NULL;
	
	// hash table set up
	hcreate(MAXFUNCNUM);

	FILE* input = fopen(INPUT_FILE,"r");

	if(argc < 3)
	{
		printf("----- Without any input ------\n");
	}
	else {
		option = argv[1];



		switch(option[0]){
			case 'm':
				if(start_analysis(input) != -1)
				{
        				strncpy(FUNC_NAMES, argv[2], QMAXPATH);
					
					FILE* finput = fopen(FUNC_NAMES, "r");
					FILE* output = fopen(OUTPUT_FILE, "w+");

					print_libsyscall_mapping(finput, output);
					fclose(finput);
					fclose(output);

				}	
				break;
			case 't':
				if(start_analysis(input) != -1)
				{
					print_test_hashtable();
				}	
				break;
			case 's':
				if(start_analysis(input) != -1)
				{
        				strncpy(TARGET_FILE, argv[2], QMAXPATH);

					//printf("The target file: %s \n", TARGET_FILE);
        
					FILE* target = fopen(TARGET_FILE, "r");

					print_target_syscall_distribution(target);
					fclose(target);
					//fclose(syscalls);
				}	
				break;
			case 'h':
			default:
				printf("------ the useful options -----\n");
				printf("  m :  generate the maping database file between glibc function and system calls-----\n");
				printf("  t :  run the test example -----\n");
				printf("  s :  generate the system calls statictis information of current program----\n");
				return 0;
		}
	}

	fclose(input);

	hdestroy();
	return 0;
}
