#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <fpro.h>
#include <allins.hpp>
#include <strlist.hpp>

int IDAP_init(void)
{
	// Do checks here to ensure your plug-in is being used within
	// an environment it was written for. Return PLUGIN_SKIP if the 	
	// checks fail, otherwise return PLUGIN_KEEP.

	//const char* options = get_plugin_options("manalysis");
	//if ( options != NULL)
	//msg("command line options: %s\n", options);	
	//qprintf("command line options: %s\n", options);
	
	return PLUGIN_KEEP;
}

void IDAP_term(void)
{
	// Stuff to do when exiting, generally you'd put any sort
	// of clean-up jobs here.
	return;
}

// The plugin can be passed an integer argument from the plugins.cfg
// file. This can be useful when you want the one plug-in to do
// something different depending on the hot-key pressed or menu
// item selected.


/* because the system call instructions are like the old days
   so this function will work every well*/
void analyze_local_syscall(FILE* result)
{
	//msg("--1. analyze the local system calls --------\n");
	int x = 0;
	int y = 0;

	for (; x < get_segm_qty(); x++)
	{
		segment_t* seg = getnseg(x);
		if(seg->type == SEG_CODE)
		{
			y = 0;
			for(; y < get_func_qty(); y++)
			{
				func_t* f = getn_func(y);
				char funcName[MAXSTR];
				get_func_name(f->startEA, funcName, sizeof(funcName) -1 );

				// enumate every instructions in the function to locate the
				// the syscall instructions
				for(ea_t addr = f->startEA; addr < f->endEA; addr++)
				{
					flags_t flags = getFlags(addr);
					
					if( isHead(flags) && isCode(flags))
					{
						char mnem[MAXSTR];
						ua_mnem(addr, mnem, sizeof(mnem) - 1);
						
						if(cmd.itype == NN_syscall)
						{
							//msg("%s found %s at %a\n",funcName, mnem, addr);
							qfprintf(result,"%s %s %a\n",funcName,mnem, addr);
						}
					}
				}
			}
		}
	}

	//msg("------ end --------\n");
}

void analyze_library_calls(FILE* result_ext, FILE* result_all)
{
	//msg("--2. analyze the import library calls ------\n");

	int i = 0;

	for(; i < get_segm_qty(); i++)
	{
		segment_t* seg = getnseg(i);
	        // summarize the function names from the code segmentation
	 	if(seg->type == SEG_CODE)
		{
			int j = 0;
			for(; j < get_func_qty(); j++)
			{
				func_t* f = getn_func(j);
				char funcName[MAXSTR] = {"0"};

				get_func_name(f->startEA, funcName, sizeof(funcName)-1);
				
				//if(funcName != NULL && f->flags == FUNC_LIB)
				if(funcName != NULL)
				{
				    // qfwrite(result,funcName,sizeof(funcName));
				    //msg("%s\n",funcName);
				    qfprintf(result_all,"%s\n",funcName);
				}	
			}
		}
	
		// get the library list from import list
		int k = 0;
		if(seg->type == SEG_XTRN) 
		{
			for(; k < get_func_qty(); k++)
			{
				func_t* f = getn_func(k);
				char funcName[MAXSTR] = {"0"};
				get_func_name(f->startEA, funcName, sizeof(funcName)-1);

				if(funcName != NULL)
				{
				    //msg("%s\n",funcName);
				    qfprintf(result_ext,"%s\n",funcName);
				}
			}
		}
	}
			
}

#define MAXLIBS 10

void IDAP_run(int arg)
{
	// The "meat" of your plug-in
	//msg("----- Recursively load all the .so and dll ------\n");
	//char loadLibs[MAXLIBS][MAXSTR];
	//int libno = 0, i;

	// traverse all the string in the symbol table to get all the dll file names
	//msg("The str number : %d\n", get_strlist_qty());

	/*for(i = 0; i < get_strlist_qty(); i++)
	{
		char internal_str[MAXSTR];
		string_info_t si;
		
		// initialize the target string 
		get_strlist_item(i, &si);

		msg("The size  : %d, %d\n", si.length, sizeof(internal_str));

		if(si.length < sizeof(internal_str))
		{
			//retrieve the string from the binary
			get_many_bytes(si.ea, internal_str, si.length);
			// C strings, 0 for C, 3 for Unicode

			msg("Str: %s\n", internal_str);
			if( si.type == 0)
			{
				// may be also .dll, match the suffix of library name to .so
				if(stristr(internal_str, ".so") && libno < MAXLIBS)
				{
					// store the library names to the array

					qstrncpy(loadLibs[libno++], internal_str, MAXSTR-1);
				} 
			}
		}
	}*/

	// load all the .so or dll libraries
	//for( i = 0; i < MAXLIBS; i++)
	//{
		//msg("Lib: %s\n", loadLibs[i]);
		
		// get the full path to the .so by searching file name.
		/*char* file = askfile_cv(0, loadLibs[i], "File path...\n", NULL);
		
		if (load_loader_module(NULL, "elf", file, 0))
		{
			msg("Successfully loaded %s\n", loadLibs[i]);
		} else {
			msg("Failed to load %s\n", loadLibs[i]);
		}*/
	//}
        //msg("----- End of the recursive analysis------\n");


	//msg("----- Start Malware Statistic Analysis------\n");


  	char allFileName[QMAXPATH] = {0};
  	char extFileName[QMAXPATH] = {0};

  	get_input_file_path(allFileName, QMAXPATH);
  	get_input_file_path(extFileName, QMAXPATH);
	int originalSize = strlen(allFileName);

  //char* index = qstrstr(targetFileName, ".idb");
  // these wild code transfer a file named as "xxx.idt" to "xxx_res" 
  	allFileName[originalSize] = '_';
  	allFileName[originalSize+1] = 'a';
  	allFileName[originalSize+2] = 'l';
  	allFileName[originalSize+3] = 'l';
  	allFileName[originalSize+4] = '\0';

  	extFileName[originalSize] = '_';
  	extFileName[originalSize+1] = 'e';
  	extFileName[originalSize+2] = 'x';
  	extFileName[originalSize+3] = 't';
  	extFileName[originalSize+4] = '\0';

        FILE* result_all = qfopen(allFileName,"w+");
        FILE* result_ext = qfopen(extFileName,"w+");

	//analyze_local_syscall(result);
	//qprintf("I'm running~~~~~~\n");
	analyze_library_calls(result_ext, result_all);

	qfclose(result_ext);
	qfclose(result_all);
	//msg("----- End of Malware Statistic Analysis------\n");
	return;
}

// There isn't much use for these yet, but I set them anyway.
char IDAP_comment[] 	= "This is my test plug-in";
char IDAP_help[] 		= "Malware Analysis";

// The name of the plug-in displayed in the Edit->Plugins menu. It can 
// be overridden in the user's plugins.cfg file.
char IDAP_name[] 		= "manalysis";

// The hot-key the user can use to run your plug-in.
char IDAP_hotkey[] 	= "Alt-X";

// The all-important exported PLUGIN object
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,	// IDA version plug-in is written for
  0,					// Flags (see below)
  IDAP_init,			// Initialisation function
  IDAP_term,			// Clean-up function
  IDAP_run,				// Main plug-in body
  IDAP_comment,			// Comment – unused
  IDAP_help,			// As above – unused
  IDAP_name,			// Plug-in name shown in 
						// Edit->Plugins menu
  IDAP_hotkey			// Hot key to run the plug-in
};

